from __future__ import annotations

import socket
import threading
import time
from typing import Callable
from typing import Optional

from .extension import KCP

SyncDataHandler = Callable[[bytes], None]


class KCPClientSync:
    def __init__(
        self,
        address: str,
        port: int,
        conv_id: int,
        no_delay: bool,
        update_interval: int,
        resend_count: int,
        no_congestion_control: bool,
        receive_window_size: int,
        send_window_size: int,
    ) -> None:
        """Configures a synchronous KCP client.

        :param address: The address of the target server.
        :param port: The port of the target server.
        :param conv_id: The conversation ID. Must be equal on both ends of the
            connection.
        :param no_delay: Whether to enable no-delay mode.
        :param update_interval: The internal update interval in milliseconds.
        :param resend_count: The number of times to resend a packet if it is
            not acknowledged.
        :param no_congestion_control: Whether to disable congestion control.
        :param receive_window_size: The size of the receive window in packets.
        :param send_window_size: The size of the send window in packets.
        """

        self.address = address
        self.port = port
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        self._kcp = KCP(
            conv_id,
            no_delay=no_delay,
            update_interval=update_interval,
            resend_count=resend_count,
            no_congestion_control=no_congestion_control,
            receive_window_size=receive_window_size,
            send_window_size=send_window_size,
        )
        self._kcp.include_outbound_handler(self._send_kcp)
        # Port 0 means the OS will pick a random port.
        self._local_port = 0

        self._data_sent = False

        # Event handlers
        self._handler: Optional[SyncDataHandler] = None
        self._on_start: Optional[Callable[[], None]] = None

    # Private API
    def _handle_data(self, data: bytes) -> None:
        if self._handler is None:
            raise RuntimeError("No data handler was set.")

        self._handler(data)

    def _send_raw_data(self, data: bytes) -> None:
        self._sock.sendto(
            data,
            (self.address, self.port),
        )
        self._data_sent = True

    def _send_kcp(self, _, data: bytes) -> None:
        self._send_raw_data(data)

    def _receive(self, data: bytes) -> None:
        self._kcp.receive(data)

        # Handle data
        for data in self._kcp.get_all_received():
            self._handle_data(data)

    def _wait_for_address(self) -> None:
        """Waits for the socket to be bound to an address."""

        while not self._data_sent:
            time.sleep(0.1)

    # Public API
    def send(self, data: bytes) -> None:
        """Enqueues data to be sent to the server on the next update cycle."""
        self._kcp.enqueue(data)

    def update_loop(self) -> None:
        """Creates a loop that updates the KCP connection."""

        while True:
            self._kcp.update()
            time.sleep(self._kcp.update_check() / 1000)

    def receive_loop(self) -> None:
        """Creates a loop that receives data from the server."""

        self._wait_for_address()

        while True:
            data, _ = self._sock.recvfrom(2048)
            self._receive(data)

    def start(self) -> None:
        """Starts the KCP Client using threads."""

        threads = [
            threading.Thread(target=self.receive_loop),
        ]

        if self._on_start is not None:
            threads.append(threading.Thread(target=self._on_start))

        for t in threads:
            t.daemon = True
            t.start()

        # Update loop is in Python, meaning signals are handled.
        self.update_loop()

    # Decorators
    def on_data(self, handler: SyncDataHandler) -> SyncDataHandler:
        """Registers a data handler to be called when data is received."""

        self._handler = handler
        return handler

    def on_start(self, handler: Callable[[], None]) -> Callable[[], None]:
        """Registers a handler to be called when the client starts."""

        self._on_start = handler
        return handler



from __future__ import annotations


class KCPException(Exception):
    """A generic exception for a KCP connection."""

    pass


class KCPConvMismatchError(KCPException):
    """Raised when the conversation ID of a packet does not match the
    conversation ID of the connection."""

    pass



from __future__ import annotations

from os import getenv
from pathlib import Path
from shutil import copyfile as copy_file

from entrypoint import entrypoint
from setuptools import Distribution, Extension
from setuptools.command.build_ext import build_ext  # type: ignore


_TRUE_SET = frozenset(("1", "true", "t", "yes", "y"))
_FALSE_SET = frozenset(("0", "false", "f", "no", "n"))


def parse_bool(string: str) -> bool:
    string = string.casefold()
    if string in _TRUE_SET:
        return True
    if string in _FALSE_SET:
        return False
    raise ValueError(f"Can't parse {string!r} into `bool`")


# Decide whether to build C extensions
EXTENSIONS = getenv("KCP_PY_EXTENSIONS")
with_extensions = True if EXTENSIONS is None else parse_bool(EXTENSIONS)

extensions = []

if with_extensions:
    from Cython.Build import cythonize  # type: ignore

    # Define your extension module(s). For Cython 3, you can use compiler_directives.
    ext = Extension(
        "kcp.extension",
        sources=["kcp/extension.pyx", "kcp/ikcp.c"],
        extra_compile_args=["-O3"],
    )

    extensions += cythonize(
        [ext],
        compiler_directives={
            "language_level": 3,
            "binding": True,
            "embedsignature": True,
            "cdivision": True,
            "wraparound": False,
            "boundscheck": False,
        },
    )


PACKAGE = "kcp"
DIRECTORY = "kcp"


def build() -> None:
    """Build the extension modules and copy them back into the package."""
    distribution = Distribution(
        {
            "name": PACKAGE,
            "ext_modules": extensions,
        }
    )
    command = build_ext(distribution)
    command.ensure_finalized()  # type: ignore
    command.run()

    # Copy built extensions back to the project (so they're importable directly)
    for output in map(Path, command.get_outputs()):  # type: ignore
        if not output.exists():
            continue
        relative_extension = output.relative_to(command.build_lib)
        copy_file(output, relative_extension)


@entrypoint(__name__)
def main() -> None:
    build()


"""Python bindings and abstarctions over the KCP protocol."""
from __future__ import annotations

__description__ = "Python bindings and abstarctions over the KCP protocol."
__version__ = "0.1.6"
__author__ = "RealistikDash <realistikdash@gmail.com>"
__license__ = "MIT"

from .client import KCPClientSync
from .server import Connection
from .server import KCPServerAsync
from .extension import KCP



from libc.stdint cimport *
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython cimport bool

import time

from .exceptions import *

cdef extern from "ikcp.h":
    # Structures
    struct IQUEUEHEAD:
        # Contents are unnecessary for our use case
        pass

    # KCP Control Object
    struct IKCPCB:
        uint32_t conv # Used to verify data is from the correct connection
        uint32_t mtu
        uint32_t mss
        uint32_t state
        uint32_t snd_una
        uint32_t snd_nxt
        uint32_t rcv_nxt
        uint32_t ts_recent
        uint32_t ts_lastack
        uint32_t ssthresh
        int32_t rx_rttval
        int32_t rx_srtt
        int32_t rx_rto
        int32_t rx_minrto
        uint32_t snd_wnd
        uint32_t rcv_wnd
        uint32_t rmt_wnd
        uint32_t cwnd
        uint32_t probe
        uint32_t current
        uint32_t interval
        uint32_t ts_flush
        uint32_t xmit
        uint32_t nrcv_buf
        uint32_t nsnd_buf
        uint32_t nrcv_que
        uint32_t nsnd_que
        uint32_t nodelay
        uint32_t updated
        uint32_t ts_probe
        uint32_t probe_wait
        uint32_t dead_link
        uint32_t incr
        IQUEUEHEAD snd_queue
        IQUEUEHEAD rcv_queue
        IQUEUEHEAD snd_buf
        IQUEUEHEAD rcv_buf
        uint32_t* acklist
        uint32_t ackcount
        uint32_t ackblock
        void* user # Just used to identify the connection it seems?
        char* buffer
        int32_t fastresend
        int32_t fastlimit
        int32_t nocwnd
        int32_t stream
        int32_t logmask
        int32_t (*output)(const char* buf, int len, IKCPCB* kcp, void *user)
        void (*writelog)(const char* log, IKCPCB* kcp, void* user);

    # Functions

    # Create KCP Control Object. User is passed to output callback.
    IKCPCB* ikcp_create(uint32_t conv, void* user)

    # Release KCP Control Object
    void ikcp_release(IKCPCB *kcp)

    # Sets the output callback function for KCP
    void ikcp_setoutput(IKCPCB *kcp, int32_t (*output)(const char* buf, int32_t len, IKCPCB* kcp, void* user))

    # (User API) Handles received KCP data
    int32_t ikcp_recv(IKCPCB *kcp, char* buffer, int32_t len)

    # (User API) Send KCP data
    int32_t ikcp_send(IKCPCB *kcp, const char* buffer, int32_t len)

    # Update KCP timing.
    void ikcp_update(IKCPCB *kcp, uint32_t current)

    # Gives time till next ikcp_update() call in ms.
    uint32_t ikcp_check(IKCPCB *kcp, uint32_t current)

    # Low level UDP packet input, call it when you receive a low level UDP packet
    int32_t ikcp_input(IKCPCB *kcp, const char *data, long size)

    # Flush pending data
    void ikcp_flush(IKCPCB *kcp)

    # Checks the size of the next message in the recv queue
    int32_t ikcp_peeksize(IKCPCB *kcp)

    # Sets the MTU for KCP
    int32_t ikcp_setmtu(IKCPCB *kcp, int32_t mtu)

    # Checks how many packets are waiting to be sent
    int32_t ikcp_waitsnd(IKCPCB *kcp)

    # ?
    int32_t ikcp_nodelay(IKCPCB *kcp, int32_t nodelay, int32_t interval, int32_t resend, int32_t nc)

    void ikcp_wndsize(IKCPCB *kcp, int32_t sndwnd, int32_t rcvwnd)

#OutboundDataHandler = Callable[[KCP, bytes], None]

# Internally used in KCP whenever data is ready to be sent.
cdef int32_t pending_outbound_data(const char* buf, int32_t len, IKCPCB* kcp, void* user) noexcept with gil:
    cdef KCP control = <KCP>user
    control.handle_output(buf, len)

cpdef get_current_time_ms():
    # Use perf counter as it isnt affected by system time changes.
    return time.perf_counter_ns() // 1000000

cdef class Clock:
    cdef uint64_t start_time
    cdef uint64_t last_time

    def __cinit__(self):
        self.start_time = get_current_time_ms()
        self.last_time = self.start_time

    cpdef uint32_t get_time(self):
        res = get_current_time_ms() - self.start_time
        self.last_time = res
        return <uint32_t>res


cdef class KCP:
    cdef IKCPCB* kcp
    # Correctly annotating this causes a cython compiler crash LOL
    cdef _data_handler # type: Optional[OutboundDataHandler]
    cdef public identity_token
    cdef Clock _clock

    def __init__(
        self,
        int conv_id,
        int max_transmission = 1400,
        bool no_delay = False,
        int update_interval = 100,
        int resend_count = 2,
        bool no_congestion_control = False,
        int send_window_size = 32,
        int receive_window_size = 128,
        identity_token = None,
    ):
        self._data_handler = None # Set by decorator.
        # Create base KCP object, passing self as the user data to be passed to the callback.
        self.kcp = ikcp_create(
            conv_id,
            <void*>self,
        )

        ikcp_setoutput(self.kcp, pending_outbound_data)

        self.set_performance_options(
            no_delay,
            update_interval,
            resend_count,
            no_congestion_control
        )

        self.set_maximum_transmission(max_transmission)

        self.set_window_size(send_window_size, receive_window_size)

        self.identity_token = identity_token

        self._clock = Clock() # todo: use a better clock

    cdef handle_output(self, const char* buf, int32_t len):
        # Create a bytes object from the buffer.
        cdef bytes data = PyBytes_FromStringAndSize(buf, len)
        self._data_handler(self, data)

    # Setting the handler for outbound data.
    def include_outbound_handler(self, handler):
        self._data_handler = handler

    # Decorator
    def outbound_handler(self, handler):
        self.include_outbound_handler(handler)
        return handler

    # I/O functions
    cpdef enqueue(self, bytes data):
        if self._data_handler is None:
            raise KCPException(
                "No outbound handler set. Cannot enqueue data. "
                "Try using the outbound_handler decorator."
            )

        if not data:
            raise ValueError("Cannot enqueue empty data.")

        cdef int32_t length = len(data)
        cdef char* buf = <char*>data
        cdef int32_t res = ikcp_send(self.kcp, buf, length)

        # Error handling
        if res < 0:
            raise KCPException(res)

    cpdef receive(self, bytes data):
        cdef int32_t length = len(data)
        cdef char* buf = <char*>data
        cdef int32_t res = ikcp_input(self.kcp, buf, length)

        # Error handling
        if res == -1:
            raise KCPConvMismatchError(
                "The conversation ID mismatch or invalid data."
            )
        elif res < -1:
            raise KCPException(res)

    # Gets the raw data received by KCP.
    cpdef bytearray get_received(self):
        # Check if there is any data to be received.
        cdef int length = ikcp_peeksize(self.kcp)
        if length == -1:
            return bytearray()

        # Create a buffer to store the data.
        cdef buf = bytearray(length)

        # Receive the data.
        cdef int32_t res = ikcp_recv(self.kcp, buf, length)
        if res < 0:
            # Theoretically this should never happen.
            raise KCPException(res)

        return buf


    # Updates timing information for KCP, may call the outbound handler. Should be regularly called.
    cpdef update(self, ts_ms: Optional[int] = None):
        # Use python's time module if no timestamp is provided.
        if ts_ms is None:
            ts_ms = self._clock.get_time()

        ikcp_update(self.kcp, ts_ms)

    # Checks when the next update should be called (in ms).
    cpdef int update_check(self, ts_ms: Optional[int] = None):
        # Use python's time module if no timestamp is provided.
        if ts_ms is None:
            ts_ms = self._clock.get_time()

        cdef res = ikcp_check(self.kcp, ts_ms)

        # TODO: This works, but idk if its the right way.
        return res - self._clock.last_time

    # Flushes the outbound data, calling the outbound handler if there is data.
    cpdef flush(self):
        ikcp_flush(self.kcp)

    # Connection settings functions
    # Sets the size of the max packet size that can be sent.
    cpdef set_maximum_transmission(self, int max_transmission):
        ikcp_setmtu(self.kcp, max_transmission)

    # Sets performance options for KCP.
    cpdef set_performance_options(
        self,
        bool no_delay,
        int update_interval,
        int resend_count,
        bool no_congestion_control,
    ):
        ikcp_nodelay(
            self.kcp,
            <int32_t>no_delay,
            update_interval,
            resend_count,
            <int32_t>no_congestion_control
        )

    # Statistics functions
    # Returns the number of packets waiting to be sent.
    cpdef int32_t get_outbound_packets(self):
        return ikcp_waitsnd(self.kcp)

    # Returns the size of the next packet to be received.
    cpdef int32_t get_next_packet_size(self):
        return ikcp_peeksize(self.kcp)

    cpdef update_loop(self):
        # TODO: Perhaps add a way to stop the loop.
        cdef int32_t next_update
        while True:
            next_update = self.update_check()
            time.sleep(next_update / 1000)
            self.update()


    # https://github.com/skywind3000/kcp/issues/289#issuecomment-770767642
    cpdef set_window_size(self, int send, int receive):
        ikcp_wndsize(self.kcp, send, receive)


    # Properties
    @property
    def next_packet_available(self):
        return self.get_next_packet_size() != -1

    # Generators
    def get_all_received(self):
        while ikcp_peeksize(self.kcp) != -1:
            yield self.get_received()



from __future__ import annotations

import asyncio
import time
from asyncio import transports
from dataclasses import dataclass
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Optional

from .extension import KCP


DataMutator = Callable[[bytes], bytes]


@dataclass
class Connection:
    _kcp: KCP
    _server: KCPServerAsync
    address: str
    port: int
    last_active: float
    _data_mutators: list[DataMutator]

    def __post_init__(self) -> None:
        self._kcp.include_outbound_handler(self._send_kcp)

    @property
    def address_tuple(self) -> AddressType:
        return self.address, self.port

    def _perform_mutations(self, data: bytes) -> bytes:
        for mutator in self._data_mutators:
            data = mutator(data)

        return data

    def __update_activity(self) -> None:
        self.last_active = time.perf_counter()

    async def receive(self, data: bytes) -> None:
        """Handles receiving data from the client."""
        self._kcp.receive(data)

        self.__update_activity()
        assert self._server._data_handler is not None  # SHUT UP MYPY

        for data in self._kcp.get_all_received():
            data = self._perform_mutations(data)
            await self._server._data_handler(self, data)

    def enqueue(self, data: bytes) -> None:
        """Enqueues data to be sent to the client."""
        self._kcp.enqueue(data)
        self.__update_activity()

    # Functions for the kcp extension
    def _send_kcp(self, _, data: bytes) -> None:
        self._server._transport.sendto(  # type: ignore
            data,
            self.address_tuple,
        )

    def update(self, ts_ms: Optional[int] = None) -> None:
        """Updates the timing information for the connection. May cause data
        to be sent to the client.

        :param ts_ms: The current time in milliseconds. If not provided, the
            the time will be calculated.
        """
        self._kcp.update(ts_ms)

    # Decorators
    def add_data_mutator(self, mutator: DataMutator) -> None:
        """Adds a local data mutator to the connection.

        :param mutator: The mutator to add.
        """

        self._data_mutators.append(mutator)


AddressType = tuple[Any, int]
DataHandler = Callable[[Connection, bytes], Awaitable[None]]
EventHandler = Callable[[], Awaitable[None]]


class KCPServerAsync(asyncio.DatagramProtocol):
    def __init__(
        self,
        address: str,
        port: int,
        conv_id: int,
        delay: int = 100,
        connection_timeout: int = 600,
        max_transmission: int = 1400,
        no_delay: bool = False,
        resend_count: int = 2,
        no_congestion_control: bool = False,
        receive_window_size: int = 128,
        send_window_size: int = 32,
    ) -> None:
        """Configures the asynchronous KCP server.

        :param address: The address to listen on (usually `127.0.0.1`).
        :param port: The port to listen on.
        :param conv: The conversation number to use for the KCP protocol.
        :param delay: The delay between KCP updates in milliseconds.
        :param connection_timeout: The amount of time in seconds to wait before
            cleaning up after a connection.
        :param max_transmission: The maximum transmission unit (MTU) of outbound
            packets.
        :param no_delay: Whether to enable no-delay mode.
        :param resend_count: The number of times to resend a packet if it is
            not acknowledged.
        :param no_congestion_control: Whether to disable congestion control.
        :param receive_window_size: The size of the receive window in packets.
        :param send_window_size: The size of the send window in packets.
        """

        self.address = address
        self.port = port
        self._transport: Optional[transports.DatagramTransport] = None
        self._loop = asyncio.get_event_loop()
        self._conv = conv_id
        self._delay = delay

        self._connections: dict[AddressType, Connection] = {}
        self.connection_timeout = connection_timeout
        self._closed = False

        # Event handlers
        self._data_handler: Optional[DataHandler] = None
        self._on_start: Optional[EventHandler] = None
        self._on_stop: Optional[EventHandler] = None

        self._data_mutators: list[DataMutator] = []

        # Connection settings.
        self._max_transmission = max_transmission
        self._no_delay = no_delay
        self._resend_count = resend_count
        self._no_congestion_control = no_congestion_control
        self._send_window_size = send_window_size
        self._receive_window_size = receive_window_size

    # Private API
    # Called by the protocol.
    def datagram_received(self, data: bytes, address: AddressType) -> None:
        connection = self._ensure_connection(address)
        self._loop.create_task(connection.receive(data))

    def _ensure_connection(self, address: AddressType) -> Connection:
        connection = self._connections.get(address)
        if connection is None:
            kcp = KCP(
                self._conv,
                max_transmission=self._max_transmission,
                no_delay=self._no_delay,
                update_interval=self._delay,
                resend_count=self._resend_count,
                no_congestion_control=self._no_congestion_control,
                send_window_size=self._send_window_size,
                receive_window_size=self._receive_window_size,
            )
            connection = Connection(
                _kcp=kcp,
                _server=self,
                address=address[0],
                port=address[1],
                last_active=time.perf_counter(),
                _data_mutators=self._data_mutators.copy(),
            )
            self._connections[address] = connection

        return connection

    # Public API
    async def listen(self) -> None:
        """Starts listening for connections alongside the KCP update loop."""
        transport, _ = await self._loop.create_datagram_endpoint(
            lambda: self,  # type: ignore
            local_addr=(self.address, self.port),
        )
        self._transport = transport

        # Call the start event handler.
        if self._on_start is not None:
            self._loop.create_task(self._on_start())  # type: ignore

        # Update connection timing information.
        while not self._closed:
            current_time = time.perf_counter()
            current_time_ms = int(current_time * 1000)
            await asyncio.sleep(self._delay / 1000)
            # Create a copy of the connections to avoid a modifying dictionary
            # while iterating over it.
            for connection in tuple(self._connections.values()):
                connection.update(current_time_ms)

                if current_time - connection.last_active > self.connection_timeout:
                    self._connections.pop(connection.address_tuple)

        # Handle cleanup.
        self._transport.close()
        self._transport = None
        self._connections.clear()

        if self._on_stop is not None:
            await self._on_stop()

    def set_performance_options(
        self,
        no_delay: Optional[bool] = None,
        update_interval: Optional[int] = None,
        resend_count: Optional[int] = None,
        no_congestion_control: Optional[bool] = None,
        receive_window_size: Optional[int] = None,
        send_window_size: Optional[int] = None,
    ) -> None:
        """Configures the performance options for all **future** KCP connections.

        :param no_delay: Whether to enable no-delay mode.
        :param update_interval: The internal update interval in milliseconds.
        :param resend_count: The number of times to resend a packet if it is
        not acknowledged.
        :param no_congestion_control: Whether to disable congestion control.
        """

        self._no_delay = no_delay if no_delay is not None else self._no_delay
        self._delay = update_interval if update_interval is not None else self._delay
        self._resend_count = resend_count if resend_count is not None else self._resend_count
        self._no_congestion_control = (
            no_congestion_control
            if no_congestion_control is not None
            else self._no_congestion_control
        )
        self._receive_window_size = (
            receive_window_size if receive_window_size is not None else self._receive_window_size
        )
        self._send_window_size = (
            send_window_size if send_window_size is not None else self._send_window_size
        )

    def start(self) -> None:
        """Creates the event loop and starts listening for connections."""
        self._loop.run_until_complete(self.listen())

    def stop(self) -> None:
        """Tells the server to stop listening for connections after the next
        update loop."""
        self._closed = True

    # Decorators
    def on_data(self, handler: DataHandler) -> DataHandler:
        """Decorator registering a handler which will be called when data is
        received from a client."""
        self._data_handler = handler
        return handler

    def on_start(self, handler: EventHandler) -> EventHandler:
        """Decorator registering a handler which will be called when the server
        starts listening for connections."""
        self._on_start = handler
        return handler

    def on_stop(self, handler: EventHandler) -> EventHandler:
        """Decorator registering a handler which will be called when the server
        stops listening for connections."""
        self._on_stop = handler
        return handler

    def add_data_mutator(self, mutator: DataMutator) -> DataMutator:
        """Decorator registering a global mutator which will be called on all data
        received from a client."""
        self._data_mutators.append(mutator)
        return mutator



from typing import Optional
from typing import Callable
from typing import Any
from typing import Generator

class KCP:
    identity_token: Any
    # Dunders
    def __init__(
        self,
        conv_id: int,
        max_transmission: int = 1400,
        no_delay: bool = True,
        update_interval: int = 100,
        resend_count: int = 2,
        no_congestion_control: bool = False,
        send_window_size: int = 32,
        receive_window_size: int = 128,
        identity_token: Any = None,
    ) -> None:
        """Creates an instance of the KCP protocol.

        :param conv_id: The conversation ID. Must be equal on both ends of the
        connection.
        :param max_transmission: The maximum transmission unit (MTU) of outbound
        packets.
        :param no_delay: Whether to enable no-delay mode.
        :param update_interval: The internal update interval in milliseconds.
        :param resend_count: The number of times to resend a packet if it is
        not acknowledged.
        :param no_congestion_control: Whether to disable congestion control.
        :param identity_token: Any data that can be used to identify the
        connection. This is not used by KCP itself.
        """
        ...
    # Handler configuration
    def include_outbound_handler(self, handler: OutboundDataHandler) -> None:
        """Adds a handler to be called when outbound data is ready to be
        sent."""
        ...
    def outbound_handler(self, handler: OutboundDataHandler) -> None:
        """Decorator equivalent of `include_outbound_handler`."""
        ...
    # I/O
    def enqueue(self, data: bytes) -> None:
        """Enqueues raw data to be sent."""
        ...
    def receive(self, data: bytes) -> None:
        """Handles receiving KCP data and adds it to the internal buffer."""
        ...
    def get_received(self) -> bytearray:
        """Returns the next received packet of data."""
        ...
    def update(self, ts_ms: Optional[int] = None) -> None:
        """Updates the connection timing information, potentially calling
        the outbound handler. This should be regularly called (~10-100ms).

        :param ts_ms: The current time in milliseconds. If not provided,
        it will be computed by the Python interpreter.
        """
        ...
    def update_check(self, ts_ms: Optional[int] = None) -> bool:
        """Checks when the next update should be called.

        :param ts_ms: The current time in milliseconds. If not provided,
        it will be computed by the Python interpreter.
        """
        ...
    def flush(self) -> None:
        """Flushes the internal buffer to the outbound handler."""
        ...
    # Configuration
    def set_maximum_transmission(self, max_transmission: int) -> None:
        """Sets the MTU (maximum transmission unit)."""
        ...
    def set_performance_options(
        self,
        no_delay: bool,
        update_interval: int,
        resend_count: int,
        no_congestion_control: bool,
    ) -> None:
        """Sets the performance options for the connection.

        :param no_delay: Whether to enable no-delay mode.
        :param update_interval: The internal update interval in milliseconds.
        :param resend_count: The number of times to resend a packet if it is
        not acknowledged.
        :param no_congestion_control: Whether to disable congestion control.
        """
        ...
    def set_window_size(self, send: int, receive: int) -> None:
        """Sets the window (buffer) size for the connection.

        :param send: The number of packets to send bnefore waiting for an
        acknowledgement.
        :param receive: The buffer size for received packets.
        """
        ...
    # Statistics
    def get_outbound_packets(self) -> int:
        """Returns the number of packets queued to be sent."""
        ...
    def get_next_packet_size(self) -> int:
        """Returns the size of the next packet received."""
        ...
    # Other
    def update_loop(self) -> None:
        """A blocking loop that continuously updates the connection
        according to `update_check()`."""
        ...
    # Properties
    @property
    def next_packet_available(self) -> bool:
        """Returns whether there is a packet available to be received."""
        ...
    # Generators
    def get_all_received(self) -> Generator[bytearray, None, None]:
        """Returns a generator that yields all received packets of data."""
        ...

OutboundDataHandler = Callable[[KCP, bytes], None]
